

//Ported from Fabio Varesano Ardunio Freeimu
//http://www.varesano.net/projects/hardware/FreeIMU#0.3.5_MS 
//=====================================================================================================
// AHRS.c
// S.O.H. Madgwick
// 25th August 2010
//=====================================================================================================
// Description:
//
// Quaternion implementation of the 'DCM filter' [Mayhony et al].  Incorporates the magnetic distortion
// compensation algorithms from my filter [Madgwick] which eliminates the need for a reference
// direction of flux (bx bz) to be predefined and limits the effect of magnetic distortions to yaw
// axis only.
//
// User must define 'halfT' as the (sample period / 2), and the filter gains 'Kp' and 'Ki'.
//
// Global variables 'q0', 'q1', 'q2', 'q3' are the quaternion elements representing the estimated
// orientation.  See my report for an overview of the use of quaternions in this application.
//
// User must call 'AHRSupdate()' every sample period and parse calibrated gyroscope ('gx', 'gy', 'gz'),
// accelerometer ('ax', 'ay', 'ay') and magnetometer ('mx', 'my', 'mz') data.  Gyroscope units are
// radians/second, accelerometer and magnetometer units are irrelevant as the vector is normalised.
//
//=====================================================================================================

//----------------------------------------------------------------------------------------------------
// Header files

#include "AHRS.h"
#include "mbed.h"

//----------------------------------------------------------------------------------------------------
// Definitions

#define Kp 3.0           // proportional gain governs rate of convergence to accelerometer/magnetometer
#define Ki 0.005        // integral gain governs rate of convergence of gyroscope biases
#define halfT 0.5        // half the sample period

//---------------------------------------------------------------------------------------------------
// Variable definitions
AHRS::AHRS(float rate) {

    firstUpdate = 0;

// quaternion elements representing the estimated orientation
    q0 = 1.0;
    q1 = 0.0;
    q2 = 0.0;
    q3 = 0.0;

// scaled integral error
    exInt = 0.0;
    eyInt = 0.0;
    ezInt = 0.0;

    deltat = rate * halfT;
}

//====================================================================================================
// Function
//====================================================================================================

void AHRS::AHRSupdate(float g_x, float g_y, float g_z, float a_x, float a_y, float a_z, float mx, float my, float mz) {

    float norm;
    float hx, hy, hz, bx, bz;
    float vx, vy, vz, wx, wy, wz;
    float ex, ey, ez;

    // auxiliary variables to reduce number of repeated operations
    
    float q0q0 = q0*q0;
    float q0q1 = q0*q1;
    float q0q2 = q0*q2;
    float q0q3 = q0*q3;
    float q1q1 = q1*q1;
    float q1q2 = q1*q2;
    float q1q3 = q1*q3;
    float q2q2 = q2*q2;   
    float q2q3 = q2*q3;
    float q3q3 = q3*q3;
    

    // normalise the measurements    
    norm = sqrt((a_x * a_x) + (a_y * a_y) + (a_z * a_z));
    a_x /= norm;
    a_y /= norm;
    a_z /= norm;
    
  
   // norm = invSqrt(a_x*a_x + a_y*a_y + a_z*a_z);
   // a_x = a_x * norm;
   // a_y = a_y * norm;
   // a_z = a_z * norm;
    
    
    norm = sqrt((mx * mx) + (my * my) + (mz * mz));
    mx /= norm;
    my /= norm;
    mz /= norm;
    
    
   // norm = invSqrt(mx*mx + my*my + mz*mz);
   // mx = mx * norm;
   // my = my * norm;
   // mz = mz * norm;
    

    // compute reference direction of flux
    hx = 2.0*mx*(0.5 - q2q2 - q3q3) + 2.0*my*(q1q2 - q0q3) + 2.0*mz*(q1q3 + q0q2);
    hy = 2.0*mx*(q1q2 + q0q3) + 2.0*my*(0.5 - q1q1 - q3q3) + 2.0*mz*(q2q3 - q0q1);
    hz = 2.0*mx*(q1q3 - q0q2) + 2.0*my*(q2q3 + q0q1) + 2.0*mz*(0.5 - q1q1 - q2q2);         
    bx = sqrt((hx*hx) + (hy*hy));
    bz = hz;
      
    // estimated direction of gravity and flux (v and w)
    vx = 2.0*(q1q3 - q0q2);
    vy = 2.0*(q0q1 + q2q3);
    vz = q0q0 - q1q1 - q2q2 + q3q3;
    wx = 2.0*bx*(0.5 - q2q2 - q3q3) + 2.0*bz*(q1q3 - q0q2);
    wy = 2.0*bx*(q1q2 - q0q3) + 2.0*bz*(q0q1 + q2q3);
    wz = 2.0*bx*(q0q2 + q1q3) + 2.0*bz*(0.5 - q1q1 - q2q2); 
    
    // error is sum of cross product between reference direction of fields and direction measured by sensors
    ex = (a_y*vz - a_z*vy) + (my*wz - mz*wy);
    ey = (a_z*vx - a_x*vz) + (mz*wx - mx*wz);
    ez = (a_x*vy - a_y*vx) + (mx*wy - my*wx);
        
    // integral error scaled integral gain
    exInt += ex*Ki;
    eyInt += ey*Ki;
    ezInt += ez*Ki;
    
    // adjusted gyroscope measurements
    g_x += Kp*ex + exInt;
    g_y += Kp*ey + eyInt;
    g_z += Kp*ez + ezInt;
    

    // integrate quaternion rate and normalise
    q0 += (-q1*g_x - q2*g_y - q3*g_z)*deltat;
    q1 += (q0*g_x + q2*g_z - q3*g_y)*deltat;
    q2 += (q0*g_y - q1*g_z + q3*g_x)*deltat;
    q3 += (q0*g_z + q1*g_y - q2*g_x)*deltat; 

    // normalise quaternion
    norm = sqrt((q0 * q0) + (q1 * q1) + (q2 * q2) + (q3 * q3));
    q0 /= norm;
    q1 /= norm;
    q2 /= norm;
    q3 /= norm;

}


void AHRS::computeEuler(float *angles) {
  
  angles[0] = atan2((2.0 * q1 * q2) - (2.0 * q0 * q3), ((2.0 * q0 * q0) - 1.0) + (2.0 * q1 * q1)); // psi
  angles[1] = -asin((2.0 * q1 * q3) + (2.0 * q0 * q2)); // theta
  angles[2] = atan2((2.0 * q2 * q3) - (2.0 * q0 * q1), ((2.0 * q0 * q0) - 1.0) + (2.0 * q3 * q3)); // phi
   
}

void AHRS::getYawPitchRoll(float *ypr) {
 
  float grx, gry, grz; // estimated gravity direction

  grx = 2.0 * q1 * q3 - (q0 * q2);
  gry = 2.0 * q0 * q1 + (q2 * q3);
  grz = (q0 * q0) - (q1 * q1) - (q2 * q2) + (q3 * q3);
  
  ypr[0] = atan2(((2.0 * q1 * q2) - (2.0 * q0 * q3)), ((2.0 * q0 * q0 - 1.0) + (2.0 * q1 * q1)));// * 180/PI;
  ypr[1] = atan(grx / sqrt(gry*gry + grz*grz));//  * 180/PI;
  ypr[2] = atan(gry / sqrt(grx*grx + grz*grz));//  * 180/PI;
}

float invSqrt(float number) {
  volatile int i;
  volatile float x, y;
  volatile const float f = 1.5F;

  x = number * 0.5F;
  y = number;
  i = * ( int * ) &y;
  i = 0x5f375a86 - ( i >> 1 );
  y = * ( float * ) &i;
  y = y * ( f - ( x * y * y ) );
  return y;
}

float AHRS::getRoll(void) {

    return phi;
   
}

float AHRS::getPitch(void) {

    return theta;
   
}

float AHRS::getYaw(void) {

    return psi;
   
}

//====================================================================================================
// END OF CODE
//====================================================================================================
